@@ -50,7 +50,7 @@
 #include <dirent.h>
 #include <errno.h>
 
-#if defined(__i386__) || defined(__x86_64__) || defined(__arm__)
+#if defined(__i386__) || defined(__x86_64__)
 #include <sys/io.h>
 #else
 #define inb(x) -1
@@ -264,6 +264,7 @@ populate_entries( struct pci_system * p )
     if (err) {
 	free(p->devices);
 	p->devices = NULL;
+	p->num_devices = 0;
     }
 
     return err;
@@ -335,9 +336,9 @@ pci_device_linux_sysfs_probe( struct pci_device * dev )
 		    dev->regions[i].size = (high_addr
 					    - dev->regions[i].base_addr) + 1;
 
-		    dev->regions[i].is_IO = (flags & 0x01);
-		    dev->regions[i].is_64 = (flags & 0x04);
-		    dev->regions[i].is_prefetchable = (flags & 0x08);
+		    dev->regions[i].is_IO = (flags & 0x01) != 0;
+		    dev->regions[i].is_64 = (flags & 0x04) != 0;
+		    dev->regions[i].is_prefetchable = (flags & 0x08) != 0;
 		}
 	    }
 
@@ -741,7 +742,7 @@ pci_device_linux_sysfs_unmap_range(struct pci_device *dev,
     return err;
 }
 
-static void pci_device_linux_sysfs_enable(struct pci_device *dev)
+static void pci_device_linux_sysfs_set_enable(struct pci_device *dev, int enable)
 {
     char name[256];
     int fd;
@@ -757,10 +758,20 @@ static void pci_device_linux_sysfs_enable(struct pci_device *dev)
     if (fd == -1)
        return;
 
-    write( fd, "1", 1 );
+    write( fd, enable ? "1" : "0" , 1 );
     close(fd);
 }
 
+static void pci_device_linux_sysfs_enable(struct pci_device *dev)
+{
+	return pci_device_linux_sysfs_set_enable(dev, 1);
+}
+
+static void pci_device_linux_sysfs_disable(struct pci_device *dev)
+{
+	return pci_device_linux_sysfs_set_enable(dev, 0);
+}
+
 static int pci_device_linux_sysfs_boot_vga(struct pci_device *dev)
 {
     char name[256];
@@ -980,7 +991,7 @@ pci_device_linux_sysfs_map_legacy(struct pci_device *dev, pciaddr_t base,
     int ret=0;
 
     if (map_flags & PCI_DEV_MAP_FLAG_WRITABLE) {
-	flags = O_RDWR; /* O_RDWR != O_WRONLY | O_RDONLY */;
+	flags = O_RDWR; /* O_RDWR != O_WRONLY | O_RDONLY */
 	prot |= PROT_WRITE;
     }
 
@@ -1041,6 +1052,7 @@ static const struct pci_system_methods linux_sysfs_methods = {
 
     .fill_capabilities = pci_fill_capabilities_generic,
     .enable = pci_device_linux_sysfs_enable,
+    .disable = pci_device_linux_sysfs_disable,
     .boot_vga = pci_device_linux_sysfs_boot_vga,
     .has_kernel_driver = pci_device_linux_sysfs_has_kernel_driver,
 
